spec JungleRewriting (*
                     This specification deals with the simplification of arithmetic expressions
                     represented by so-called jungles. A jungle is a syntax "tree" with identical
                     subtrees being identified in order not having to evaluate the same expres-
                     sion twice or even more times. As a consequence, there may be many fathers
                     of a node in a jungle and thus, a jungle is a dag (directed, acyclic graph)
                     rather than a tree.
                     There are three main sections within this specification:
                     Section "Jungle_Scheme" contains all information about static properties
                     of the class of graphs representing jungles.
                     Section "Create_Jungle" provides productions for the construction of a
                     jungle. In this version, in order to keep this specification as simple
                     as possible, there is only a single production creating a sample jungle.
                     It may easily be extended to contain productions for insertion or expansion
                     of nodes.
                     Section "Optimize_Jungle" consists of productions that simplify
                     arithmetic expressions.
                     *)
                     (* @CC_LT HoriWithIndent @*)

   declares

      section Jungle_Scheme

         declares

            node_class EXPR
               derived
                  Value : integer [0:1];
                  GRAPH_BROWSER_LABEL : string;
            end;

            node_class LITERAL is_a EXPR
               intrinsic
                  Text : string;
               redef_derived
                  GRAPH_BROWSER_LABEL = self.Text;
            end;

            node_type Const : LITERAL
               redef_derived
                  Value = value ( self.Text, integer [0:1] );
            end;

            node_type Var : LITERAL
               redef_derived
                  Value = nil;
            end;

            node_class OPERATOR is_a EXPR end;

            node_type Plus : OPERATOR
               redef_derived
                  Value = self.-Operand1->.Value + self.-Operand2->.Value;
                  GRAPH_BROWSER_LABEL = "+";
            end;

            node_type Minus : OPERATOR
               redef_derived
                  Value = self.-Operand1->.Value - self.-Operand2->.Value;
                  GRAPH_BROWSER_LABEL = "-";
            end;

            node_type Times : OPERATOR
               redef_derived
                  Value = self.-Operand1->.Value * self.-Operand2->.Value;
                  GRAPH_BROWSER_LABEL = "*";
            end;

            node_type Div : OPERATOR
               redef_derived
                  Value = self.-Operand1->.Value / self.-Operand2->.Value;
                  GRAPH_BROWSER_LABEL = "/";
            end;

            node_type Root : OPERATOR
               redef_derived
                  Value = nil;
                  GRAPH_BROWSER_LABEL = "Root";
            end;

            edge_type Operand1 : OPERATOR -> EXPR [1:1];

            edge_type Operand2 : OPERATOR -> EXPR [1:1];

            path ToOperand : OPERATOR -> EXPR [1:n] =
                   -Operand1->
               or  -Operand2->
            end;

            path DownToAll : EXPR -> EXPR [0:n] =
               (    instance_of OPERATOR
                 &  (     -Operand1->
                      or  -Operand2-> )  ) *
            end;

            function `Equiv' : ( t1, t2 : EXPR) -> boolean =
               [ (t1.type # t2.type) :: false
               | (t1.type in LITERAL) :: ((t1 : LITERAL).Text = (t2 : LITERAL).Text)
               | ((t1 : OPERATOR).-Operand1-> `Equiv' (t2 : OPERATOR).-Operand1->) and ((t1 : OPERATOR).-Operand2-> `Equiv' (t2 : OPERATOR).-Operand2->)
                  ]
            end;

      end;

      section Create_Jungle

         declares

            transformation CreateSampleGraph =
               (<OptPrecondDecl>)
               (* @CC_LG 7008(8,8)(27,33) @*)

               (<OptLeftSideList>)
               (* @CC_LG 7008C(27,33) @*)
               ::=
               begin
                  3' : Plus;
                  (* @CC_LG 6012(672,448)(97,29) @*)
                  4' : Times;
                  (* @CC_LG 6012(744,350)(97,29) @*)
                  3' -> 4' : Operand1 (* @CC_LG 1002C,BC(6,3)(48,10) @*);
                  (* @CC_LG 3004C @*)
                  5' : Times;
                  (* @CC_LG 6012(600,350)(97,29) @*)
                  3' -> 5' : Operand2 (* @CC_LG 1002(10,20)(48,10) @*);
                  (* @CC_LG 3004C @*)
                  6' : Plus;
                  (* @CC_LG 6012(912,252)(97,29) @*)
                  4' -> 6' : Operand1 (* @CC_LG 1002(-53,9)(48,10) @*);
                  (* @CC_LG 3004C @*)
                  11' : Minus;
                  (* @CC_LG 6012(624,154)(97,29) @*)
                  6' -> 11' : Operand2 (* @CC_LG 1002(11,11)(48,10) @*);
                  (* @CC_LG 3004C @*)
                  10' : Var;
                  (* @CC_LG 6012(672,56)(97,29) @*)
                  6' -> 10' : Operand1 (* @CC_LG 1002(-57,11)(48,10) @*);
                  (* @CC_LG
                  3004C4
                  864,196
                  720,112 @*)
                  7' : Plus;
                  (* @CC_LG 6012(936,154)(97,29) @*)
                  12' : Var;
                  (* @CC_LG 6012(1104,56)(97,29) @*)
                  13' : Const;
                  (* @CC_LG 6012(960,56)(97,29) @*)
                  7' -> 12' : Operand1 (* @CC_LG 1002(-41,11)(48,10) @*);
                  (* @CC_LG
                  3004C3
                  1032,112 @*)
                  7' -> 13' : Operand2 (* @CC_LG 1002(3,11)(48,10) @*);
                  (* @CC_LG
                  3004C3
                  936,112 @*)
                  4' -> 7' : Operand2 (* @CC_LG 1002C,BC(0,-3)(48,10) @*);
                  (* @CC_LG
                  3004C4
                  768,280
                  816,224 @*)
                  8' : Plus;
                  (* @CC_LG 6012(504,252)(97,29) @*)
                  9' : Var;
                  (* @CC_LG 6012(1248,56)(97,29) @*)
                  14' : Times;
                  (* @CC_LG 6012(480,154)(97,29) @*)
                  15' : Const;
                  (* @CC_LG 6012(48,56)(97,29) @*)
                  5' -> 8' : Operand1 (* @CC_LG 1002(-41,14)(48,10) @*);
                  (* @CC_LG 3004C @*)
                  5' -> 9' : Operand2 (* @CC_LG 1002(0,14)(48,10) @*);
                  (* @CC_LG
                  3004C6
                  624,308
                  624,280
                  648,252
                  864,140 @*)
                  8' -> 14' : Operand1 (* @CC_LG 1002(-48,11)(48,10) @*);
                  (* @CC_LG
                  3004C3
                  480,196 @*)
                  8' -> 15' : Operand2 (* @CC_LG 1002(0,11)(48,10) @*);
                  (* @CC_LG
                  3004C4
                  432,196
                  144,112 @*)
                  1' : Plus;
                  (* @CC_LG 6012(744,546)(97,29) @*)
                  1' -> 3' : Operand2 (* @CC_LG 1002(0,4)(48,10) @*);
                  (* @CC_LG
                  3004C3
                  696,504 @*)
                  2' : Var;
                  (* @CC_LG 6012(816,56)(97,29) @*)
                  1' -> 2' : Operand1 (* @CC_LG 1002C,BC(0,-8)(48,10) @*);
                  (* @CC_LG
                  3004C6
                  792,476
                  792,420
                  840,364
                  840,112 @*)
                  16' : Const;
                  (* @CC_LG 6012(336,56)(97,29) @*)
                  17' : Const;
                  (* @CC_LG 6012(192,56)(97,29) @*)
                  11' -> 16' : Operand1 (* @CC_LG 1002(-48,10)(48,10) @*);
                  (* @CC_LG 3004C @*)
                  11' -> 17' : Operand2 (* @CC_LG 1002(0,10)(48,10) @*);
                  (* @CC_LG 3004C @*)
                  18' : Var;
                  (* @CC_LG 6012(528,56)(97,29) @*)
                  19' : Const;
                  (* @CC_LG 6012(456,56)(97,29) @*)
                  14' -> 18' : Operand1 (* @CC_LG 1002(-41,10)(48,10) @*);
                  (* @CC_LG 3004C @*)
                  14' -> 19' : Operand2 (* @CC_LG 1002(0,10)(48,10) @*);
                  (* @CC_LG
                  3004C3
                  456,112 @*)
                  20' : Root;
                  (* @CC_LG 6012(744,644)(97,29) @*)
                  20' -> 1' : Operand1 (* @CC_LG 1002C,BC(6,22)(48,10) @*);
                  (* @CC_LG 3004C @*)
               end
               (* @CC_LG 7009(8,8)(1309,679) @*)
               transfer 2'.Text := "b";
                        9'.Text := "b";
                        10'.Text := "a";
                        12'.Text := "a";
                        13'.Text := "1";
                        15'.Text := "1";
                        16'.Text := "8";
                        17'.Text := "7";
                        18'.Text := "a";
                        19'.Text := "1";
            end;

      end;

      section GraphBrowserInitialization (* @CC_LT HoriWithSpace @*)

         declares

            transformation InitGraphBrowserLabel (* This production is needed for technical reasons:
                                                 In order to enable the graph browser to display the attribute
                                                 GRAPH_BROWSER_LABEL, this attribute has to be initialized.
                                                 Due to the incremental attribute evaluation within PROGRES,
                                                 variables aren't initialized until they are read. Therefore,
                                                 this production reads the mentioned attribute. It doesn't
                                                 change the graph in any way.
                                                 *)
                                                 (* @CC_LT HoriWithIndent @*)
               =

               begin
                  obl_set `1 : EXPR;
                  (* @CC_LG 6011(60,28)(97,21) @*)
                  => `1 : valid (self.GRAPH_BROWSER_LABEL # "");
                  (* @CC_LG
                  3011XC2
                  182,77 @*)
               end
               (* @CC_LG 7008(8,8)(423,105) @*)
               ::=
               begin
                  obl_set 1' = `1;
                  (* @CC_LG 6011(60,28)(97,21) @*)
               end
               (* @CC_LG 7009(8,8)(135,61) @*)
            end;

      end;

      section Optimize_Jungle

         declares

            transformation MergeCommonSubexpressions =

               begin
                  obl_node `1 : EXPR;
                  (* @CC_LG 6007(48,28)(71,21) @*)
                  obl_node `2 : EXPR;
                  (* @CC_LG 6007(132,28)(71,21) @*)
               end
               (* @CC_LG 7008(8,8)(185,62) @*)
               ::=
               begin
                  obl_node 1' = `1;
                  (* @CC_LG 6007(52,32)(59,21) @*)
               end
               (* @CC_LG 7009(8,8)(97,55) @*)
               condition `1 `Equiv' `2;
               embedding redirect <-Operand1-, <-Operand2- from `2 to 1';
            end;

            transformation SimplifyExprWithKnownValue (* This production simplifies expressions like
                                                      <operand1> <operator> <operand2>
                                                      with operands with a known value to a
                                                      single constant.
                                                      *)
                                                      (* @CC_LT HoriWithSpace @*)
               =

               begin
                  obl_node `1 : OPERATOR;
                  (* @CC_LG 6007(66,22)(95,21) @*)
               end
               (* @CC_LG 7008(8,8)(126,48) @*)
               ::=
               begin
                  1' : Const;
                  (* @CC_LG 6012(60,28)(97,21) @*)
               end
               (* @CC_LG 7009(8,8)(133,55) @*)
               condition not empty ( `1.Value );
               embedding redirect <-Operand1-, <-Operand2- from `1 to 1';
               transfer 1'.Text := string ( `1.Value ) : string [1:1];
            end;

            transformation SimplifyZeroPlusX (* Simplifies the expressions 0+x and x+0 to  x.
                                             Note that the node `2 representing the constant
                                             zero must not be deleted since there may be
                                             other operators pointing to it.
                                             *)
                                             (* @CC_LT HoriWithSpace @*)
               =

               begin
                  obl_node `1 : Plus;
                  (* @CC_LG 6007(105,22)(71,21) @*)
                  obl_node `2 : Const;
                  (* @CC_LG 6007(52,90)(77,21) @*)
                  obl_node `3 : EXPR;
                  (* @CC_LG 6007(159,90)(71,21) @*)
                  `1 => `2 : ToOperand;
                  (* @CC_LG 3006C @*)
                  `1 => `3 : ToOperand;
                  (* @CC_LG 3006C @*)
               end
               (* @CC_LG 7008(8,8)(216,113) @*)
               ::=
               begin
                  obl_node 3' = `3;
                  (* @CC_LG 6007(156,28)(59,21) @*)
                  obl_node 2' = `2;
                  (* @CC_LG 6007(48,28)(59,21) @*)
               end
               (* @CC_LG 7009(8,8)(205,55) @*)
               condition (`2.Value) : integer [1:1] = 0;
               embedding redirect <-Operand1-, <-Operand2- from `1 to 3';
            end;

            transformation SimplifyZeroTimesX (* Simplifies the expressions 0*x and x*0 to 0.
                                              Note that `3 must not be deleted since ther
                                              may be other operators pointing to this node.
                                              *)
                                              (* @CC_LT HoriWithSpace @*)
               =

               begin
                  obl_node `1 : Times;
                  (* @CC_LG 6007(108,22)(77,21) @*)
                  obl_node `2 : Const;
                  (* @CC_LG 6007(52,90)(77,21) @*)
                  obl_node `3 : EXPR;
                  (* @CC_LG 6007(161,90)(71,21) @*)
                  `1 => `2 : ToOperand;
                  (* @CC_LG 3006C @*)
                  `1 => `3 : ToOperand;
                  (* @CC_LG 3006C @*)
               end
               (* @CC_LG 7008(8,8)(209,121) @*)
               ::=
               begin
                  obl_node 2' = `2;
                  (* @CC_LG 6007(48,28)(59,21) @*)
                  obl_node 3' = `3;
                  (* @CC_LG 6007(156,28)(59,21) @*)
               end
               (* @CC_LG 7009(8,8)(205,55) @*)
               condition (`2.Value) : integer [1:1] = 0;
               embedding redirect <-Operand1-, <-Operand2- from `1 to 2';
            end;

            transformation SimplifyOneTimesX (* Simplifies the expressions 1*x and x*1 to x.
                                             Note that node `2 representing the constant
                                             one must not be deleted since there may
                                             exist other operators pointing to it.
                                             *)
                                             (* @CC_LT HoriWithSpace @*)
               =

               begin
                  obl_node `1 : Times;
                  (* @CC_LG 6007(105,32)(77,21) @*)
                  obl_node `2 : Const;
                  (* @CC_LG 6007(52,104)(77,21) @*)
                  obl_node `3 : EXPR;
                  (* @CC_LG 6007(159,104)(71,21) @*)
                  `1 => `2 : ToOperand;
                  (* @CC_LG 3006C @*)
                  `1 => `3 : ToOperand;
                  (* @CC_LG 3006C @*)
               end
               (* @CC_LG 7008(8,8)(216,127) @*)
               ::=
               begin
                  obl_node 3' = `3;
                  (* @CC_LG 6007(128,28)(59,21) @*)
                  obl_node 2' = `2;
                  (* @CC_LG 6007(36,28)(59,21) @*)
               end
               (* @CC_LG 7009(8,8)(169,55) @*)
               condition (`2.Value) : integer [1:1] = 1;
               embedding redirect <-Operand1-, <-Operand2- from `1 to 3';
            end;

            transformation SimplifyXMinusX (* Simplifies the expression x-x to 0.
                                           Note that `2 must not be deleted because
                                           there may exist other nodes pointing to it.
                                           *)
                                           (* @CC_LT HoriWithIndent @*)                =

               begin
                  obl_node `1 : Minus;
                  (* @CC_LG 6007(88,32)(77,21) @*)
                  obl_node `2 : EXPR;
                  (* @CC_LG 6007(88,104)(71,21) @*)
                  `1 -> `2 : Operand1 (* @CC_LG 1002C,BC(0,-26)(48,10) @*);
                  (* @CC_LG 3004C @*)
                  `1 -> `2 : Operand2 (* @CC_LG 1002C,BC(0,80)(48,10) @*);
                  (* @CC_LG 3004C @*)
               end
               (* @CC_LG 7008(8,8)(181,127) @*)
               ::=
               begin
                  3' : Const;
                  (* @CC_LG 6012(96,56)(97,29) @*)
                  obl_node 2' = `2;
                  (* @CC_LG 6007(221,84)(59,21) @*)
               end
               (* @CC_LG 7009(8,8)(263,115) @*)
               embedding redirect <-Operand1-, <-Operand2- from `1 to 3';
               transfer 3'.Text := string ( 0 );
            end;

            transformation ApplyDistributiveLaw =

               begin
                  obl_node `1 : Plus;
                  (* @CC_LG 6007(192,28)(71,21) @*)
                  obl_node `2 : Times;
                  (* @CC_LG 6007(108,90)(77,21) @*)
                  obl_node `3 : Times;
                  (* @CC_LG 6007(268,90)(77,21) @*)
                  `1 -> `2 : Operand1;
                  (* @CC_LG 3004C @*)
                  `1 -> `3 : Operand2;
                  (* @CC_LG 3004C @*)
                  obl_node `4 : EXPR;
                  (* @CC_LG 6007(56,168)(71,21) @*)
                  obl_node `6 : EXPR;
                  (* @CC_LG 6007(304,168)(71,21) @*)
                  `2 -> `4 : Operand1 (* @CC_LG 1002C,BC(10,15)(48,10) @*);
                  (* @CC_LG 3004C @*)
                  `3 -> `6 : Operand2 (* @CC_LG 1002(-9,12)(48,10) @*);
                  (* @CC_LG 3004C @*)
                  obl_node `5 : EXPR;
                  (* @CC_LG 6007(156,168)(71,21) @*)
                  `2 -> `5 : Operand2 (* @CC_LG 1002C,BC(11,29)(48,10) @*);
                  (* @CC_LG 3004C @*)
                  `3 -> `4 : Operand1 (* @CC_LG 1002(-59,12)(48,10) @*);
                  (* @CC_LG 3004C @*)
               end
               (* @CC_LG 7008(8,8)(359,200) @*)
               ::=
               begin
                  1' : Times;
                  (* @CC_LG 6011(123,30)(77,21) @*)
                  obl_node 2' = `4;
                  (* @CC_LG 6007(52,90)(59,21) @*)
                  3' : Plus;
                  (* @CC_LG 6011(176,90)(71,21) @*)
                  1' -> 2' : Operand1 (* @CC_LG 1002C,BC(0,-6)(48,10) @*);
                  (* @CC_LG 3004C @*)
                  1' -> 3' : Operand2 (* @CC_LG 1002C,BC(6,36)(48,10) @*);
                  (* @CC_LG 3004C @*)
                  obl_node 4' = `5;
                  (* @CC_LG 6007(123,154)(59,21) @*)
                  obl_node 5' = `6;
                  (* @CC_LG 6007(230,154)(59,21) @*)
                  3' -> 4' : Operand1 (* @CC_LG 1002C,BC(-4,24)(48,10) @*);
                  (* @CC_LG 3004C @*)
                  3' -> 5' : Operand2 (* @CC_LG 1002C,BC(-4,24)(48,10) @*);
                  (* @CC_LG 3004C @*)
               end
               (* @CC_LG 7009(8,8)(275,186) @*)
               folding { `4, `5, `6 };
               embedding redirect <-Operand1-, <-Operand2- from `1 to 1';
            end;

            transformation DoGarbageCollection (* Most simple form of a garbage collection:
                                               All nodes that are no longer reachable from
                                               the root are deleted.
                                               *)
                                               (* @CC_LT HoriWithSpace @*)
               =

               begin
                  obl_node `1 : Root;
                  (* @CC_LG 6007(48,150)(71,21) @*)
                  opt_set `2 : EXPR;
                  (* @CC_LG 6009(48,98)(71,21) @*)
                  `1 #> `2 : DownToAll
                             (* @CC_LG 4005C,TC(0,11)(54,12) @*);
                  (* @CC_LG 3009C @*)
               end
               (* @CC_LG 7008(8,8)(162,181) @*)
               ::=
               begin
                  obl_node 1' = `1;
                  (* @CC_LG 6007(48,28)(59,21) @*)
               end
               (* @CC_LG 7009(8,8)(99,61) @*)
               (* @CC_LT HoriWithIndent @*)
            end;

            transformation Optimize =

               loop
                  InitGraphBrowserLabel
                  & choose
                       SimplifyOneTimesX
                    else
                       SimplifyZeroPlusX
                    else
                       SimplifyZeroTimesX
                    else
                       SimplifyXMinusX
                    else
                       SimplifyExprWithKnownValue
                    else
                       MergeCommonSubexpressions
                    else
                       ApplyDistributiveLaw
                    end
                  & DoGarbageCollection
               end
            end;

      end;

      section OptimizeSampleGraph

         declares

            transformation MAIN =

               CreateSampleGraph
               & Optimize
            end;

      end;

end.
