spec LibraryInformationSystem (* This specification describes the static structure and the dynamic
                              behavior of a Library Information System. The system models different views
                              on the library:
                              - The library administrator maintains the catalogue of books in the
                              library and keeps the information about book titles and magazines,
                              e.g. keywords, up-to-date in the central database.
                              - The user administrator is able to add and remove library users, and
                              handles the lending and returning of books or magazines.
                              - The user view comprises facilities for querying books matching certain
                              search criteria.
                              *)

   declares

      external_package DateTime

         types Date, Time ;

         functions
            CurrentDate : -> Date; Days : ( Date, Date) -> integer; CurrentTime : -> Time;

      end;

      package MainPackage

         imports
            package DateTime
               use
                  function CurrentDate;
                  function Days;
                  type Date;
            end;

         declares

            section DISPLAY_BASICS

               declares

                  node_class DISPLAY_ITEM
                     intrinsic
                        icon : string := "";
                     derived
                        label : string = "";
                        visible : integer [0:n] = 0;
                  end;

                  query touch( d : DISPLAY_ITEM) =
                     begin
                        obl_node `1 = d;
                        (* @CC_LG 6007(96,56)(97,29) @*)
                        => `1 : valid (self.label # "emptyLabel");
                        (* @CC_LG
                        3011XC2
                        218,108 @*)
                     end
                  end;

                  query touchAll =
                     begin
                        opt_set `1 : DISPLAY_ITEM;
                        (* @CC_LG 6010(96,56)(97,29) @*)
                        => `1 : valid (self.label # "emptyLabel");
                        (* @CC_LG
                        3011XC2
                        218,108 @*)
                     end
                  end;

                  transformation RemoveAll =

                     begin
                        opt_set `1 : DISPLAY_ITEM;
                        (* @CC_LG 6010(96,56)(97,29) @*)
                     end
                     ::=
                  end;

            end;

            section LIBRARY_SCHEMA

               declares

                  node_class DB_OBJECT is_a DISPLAY_ITEM
                     intrinsic
                        id : integer := 0;
                  end;

                  node_class LIBRARY_INFORMATION_SYSTEM is_a DISPLAY_ITEM end;

                  node_type LIS : LIBRARY_INFORMATION_SYSTEM
                     redef_derived
                        label = string ( self.type );
                  end;

                  node_class LIBRARY is_a DB_OBJECT
                     intrinsic
                        bookLimit : integer := 10;
                        currUserId : integer := 1000;
                        currItemId : integer := 300000;
                     derived
                        noOfUsers : integer = card ( self.user );
                     redef_derived
                        label = string ( self.type );
                  end;

                  edge_type partOf : LIBRARY [0:n] -> LIBRARY_INFORMATION_SYSTEM [1:1];

                  node_type LendingLibrary : LIBRARY end;

                  node_type ReferenceLibrary : LIBRARY
                     redef_intrinsic
                        bookLimit := 0;
                  end;

                  node_type Item : DB_OBJECT
                     intrinsic
                        loanDate : Date [0:1] := nil;
                     redef_derived
                        label = string ( self.id ) & ": " & self.-has->.label & "\n" & [ self.<-owns-.bookLimit > 0 :: [ self is with <-borrows- :: "borrowed " & string ( self.loanDate ) : [1:1]
                                                                                                                       | "available" ]
                                                                                       | "not available" ]                                                                                        ;
                  end;

                  edge_type owns : LIBRARY [1:1] -> Item [0:n];

                  edge_type lost : LIBRARY [1:1] -> Item [0:n];

                  edge_type has : Item [0:n] -> TITLE [1:1];

                  node_type Borrower : DB_OBJECT
                     intrinsic
                        index name : string := "";
                     derived
                        borrowedItems : integer [0:n] = self.-borrows->.id;
                     constraint
                        BelowBookLimit = for_all Lib := self.usedLibs ::
                                            card ( self.hasItem ( Lib ) ) <= Lib.bookLimit
                                         end                                              ;
                     redef_derived
                        label = string ( self.id ) & ": " & self.name & [ self is with -borrows-> :: "\n" & ListStrings ( string ( self.borrowedItems ) ) : string [1:1]
                                                                        | "" ]                                                                                          ;
                  end;

                  edge_type borrows : Borrower [0:1] -> Item [0:n];

                  edge_type user : LIBRARY [0:n] -> Borrower [0:n];

                  function ListStrings : ( s : string [0:n]) -> string [0:1] =
                     [ not empty ( s ) :: StringConc ( "", all s ) ]
                  end;

                  function StringConc : ( s1 : string ; s2 : string) -> string =
                     [ s1 = "" :: s2
                     | s1 & ", " & s2 ]
                  end;

                  node_class TITLE is_a DISPLAY_ITEM
                     intrinsic
                        title : string := "";
                        authors : string [0:n] := nil;
                        editors : string [0:n] := nil;
                     constraint
                        NoSubCycle = not (self in self.( subtitle + ));
                     redef_derived
                        label = [ ListStrings ( self.authors ) & ":\n"
                                | ListStrings ( self.editors ) & ":\n"
                                | "" ]                                 & self.title;
                  end;

                  edge_type subtitle : TITLE [0:1] -> TITLE [0:n];

                  node_type Magazin : TITLE end;

                  node_type Series : TITLE end;

                  node_type Book : TITLE end;

            end;

            section LIB_DERIVED_RELATIONSHIPS

               declares

                  function .. : ( lb, ub : integer) -> integer [0:n] =
                     [ lb <= ub :: lb or ((lb + 1) .. ub) ]
                  end;

                  static path usedLibs : Borrower [0:n] -> LIBRARY [0:n] =
                        -borrows->
                     &  <-owns-
                  end;

                  path hasItem( In : LIBRARY) : Borrower [0:1] -> Item [0:n] =
                     `1 => `2 in
                     begin
                        obl_node `1 : Borrower;
                        (* @CC_LG 6007(216,154)(97,29) @*)
                        obl_node `2 : Item;
                        (* @CC_LG 6007(144,56)(97,29) @*)
                        obl_node `3 = In;
                        (* @CC_LG 6007(72,154)(97,29) @*)
                        `1 -> `2 : borrows;
                        (* @CC_LG 3004C @*)
                        `3 -> `2 : owns;
                        (* @CC_LG 3004C @*)
                     end
                  end;

                  restriction overdue( DayLimit : integer) : Item =
                     valid (Days ( CurrentDate, [ self.loanDate
                                                | CurrentDate ] ) > DayLimit)
                  end;

            end;

            section LIB_ADMIN

               declares

                  transformation AddTitle ( titleType : type_in TITLE ; titleName : string ; authorNames : string ; out thisTitle : TITLE)
                     [0:1] =

                     begin
                        not_node `1 : TITLE;
                        (* @CC_LG 6013(96,56)(97,29) @*)
                        => `1 : valid (self.title = titleName);
                        (* @CC_LG
                        3011XC2
                        218,108 @*)
                     end
                     ::=
                     begin
                        1' : titleType;
                        (* @CC_LG 6012(96,56)(97,29) @*)
                     end
                     transfer 1'.title := titleName;
                              1'.authors := authorNames;
                     return thisTitle := 1';
                  end;

                  transformation MakeSubtitle ( title : TITLE ; subTitle : TITLE) [0:1] =

                     begin
                        obl_node `1 = title;
                        (* @CC_LG 6007(96,56)(97,29) @*)
                        obl_node `2 = subTitle;
                        (* @CC_LG 6007(168,154)(97,29) @*)
                        `2 #> `1 : subtitle +;
                        (* @CC_LG
                        3009C3
                        120,112 @*)
                        not_node `3 : TITLE;
                        (* @CC_LG 6013(240,56)(97,29) @*)
                        `2 -> `3 : subtitle;
                        (* @CC_LG
                        3004C3
                        216,112 @*)
                     end
                     ::=
                     begin
                        obl_node 1' = `1;
                        (* @CC_LG 6007(96,154)(97,29) @*)
                        obl_node 2' = `2;
                        (* @CC_LG 6007(96,56)(97,29) @*)
                        1' -> 2' : subtitle;
                        (* @CC_LG 3004C @*)
                     end
                  end;

                  transformation AddItem ( title : TITLE ; library : LIBRARY ; out item : Item) [1:1]
                     =

                     begin
                        obl_node `1 = library;
                        (* @CC_LG 6007(96,56)(97,29) @*)
                        obl_node `2 = title;
                        (* @CC_LG 6007(240,56)(97,29) @*)
                     end
                     ::=
                     begin
                        obl_node 1' = `1;
                        (* @CC_LG 6007(96,252)(97,29) @*)
                        obl_node 2' = `2;
                        (* @CC_LG 6007(96,56)(97,29) @*)
                        3' : Item;
                        (* @CC_LG 6012(96,154)(97,29) @*)
                        1' -> 3' : owns;
                        (* @CC_LG 3004C @*)
                        3' -> 2' : has;
                        (* @CC_LG 3004C @*)
                     end
                     transfer 3'.id := `1.currItemId;
                              1'.currItemId := `1.currItemId + 1;
                              3'.loanDate := nil;
                     return item := 3';
                  end;

                  transformation RemoveItem ( item : Item) [0:1] =

                     begin
                        obl_node `1 = item;
                        (* @CC_LG 6007(96,56)(97,29) @*)
                        not_node `2 : Borrower;
                        (* @CC_LG 6013(96,154)(97,29) @*)
                        `2 -> `1 : borrows;
                        (* @CC_LG 3004C @*)
                     end
                     ::=
                  end;

                  transformation CleanDB =

                     use Name : string;
                         User : Borrower;
                         LostItems : Item [1:n]
                     do
                        loop
                           RemoveOverdueBorrower ( out Name, out LostItems )
                           & for_all i := elem ( LostItems )
                             do
                                RemoveItem ( i )
                             end
                        end
                     end
                  end;

            end;

            section LIB_USER_ADMIN

               declares

                  section LIB_BORROWERS

                     declares

                        transformation AddLibUser ( library : LIBRARY ; userName : string ; out borrower : Borrower)
                           [1:1] =

                           begin
                              obl_node `1 = library;
                              (* @CC_LG 6007(96,56)(97,29) @*)
                           end
                           ::=
                           begin
                              2' : Borrower;
                              (* @CC_LG 6012(96,56)(97,29) @*)
                              obl_node 1' = `1;
                              (* @CC_LG 6007(96,154)(97,29) @*)
                              1' -> 2' : user;
                              (* @CC_LG 3004C @*)
                           end
                           transfer 2'.name := userName;
                                    2'.id := `1.currUserId;
                                    1'.currUserId := `1.currUserId + 1;
                           return borrower := 2';
                        end;

                        transformation RemoveLibUser ( borrower : Borrower) =

                           begin
                              obl_node `1 = borrower;
                              (* @CC_LG 6007(96,154)(97,29) @*)
                              not_node `2 : Item;
                              (* @CC_LG 6013(96,56)(97,29) @*)
                              `1 -> `2 : borrows;
                              (* @CC_LG 3004C @*)
                           end
                           ::=
                        end;

                        transformation ForceRemoveBorrower ( borrower : Borrower ; out lostItems : Item [0:n])
                           =

                           begin
                              opt_set `2 : Item;
                              (* @CC_LG 6010(168,56)(97,29) @*)
                              obl_node `3 : LIBRARY;
                              (* @CC_LG 6007(96,154)(97,29) @*)
                              `3 -> `2 : owns;
                              (* @CC_LG 3004C @*)
                              obl_node `1 = borrower;
                              (* @CC_LG 6007(240,154)(97,29) @*)
                              `1 -> `2 : borrows;
                              (* @CC_LG 3004C @*)
                           end
                           ::=
                           begin
                              opt_set 2' = `2;
                              (* @CC_LG 6010(96,56)(97,29) @*)
                              obl_node 3' = `3;
                              (* @CC_LG 6007(96,154)(97,29) @*)
                              3' -> 2' : lost;
                              (* @CC_LG 3004C @*)
                           end
                           return lostItems := 2';
                        end;

                        transformation RemoveOverdueBorrower ( out userName : string ; out lostItems : Item [1:n])
                           =

                           begin
                              obl_node `1 : Borrower;
                              (* @CC_LG 6007(144,154)(97,29) @*)
                              obl_node `2 : Item;
                              (* @CC_LG 6007(240,56)(97,29) @*)
                              `1 -> `2 : borrows;
                              (* @CC_LG 3004C @*)
                              => `2 : overdue ( 100 );
                              (* @CC_LG
                              3011XC2
                              362,108 @*)
                              obl_set `3 : Item;
                              (* @CC_LG 6011(96,56)(97,29) @*)
                              `1 -> `3 : borrows;
                              (* @CC_LG 3004C @*)
                           end
                           ::=
                           begin
                              obl_node 2' = `2;
                              (* @CC_LG 6007(240,56)(97,29) @*)
                              obl_set 3' = `3;
                              (* @CC_LG 6011(96,56)(97,29) @*)
                           end
                           folding { `2, `3 };
                           return userName := `1.name;
                                  lostItems := 3';
                        end;

                  end;

                  section LIB_USER_ADMIN_QUERIES

                     declares

                        query GetOverdueItems( out items : Item [0:n]) =
                           begin
                              opt_set `1 : Item;
                              (* @CC_LG 6010(96,56)(97,29) @*)
                              => `1 : overdue ( 100 );
                              (* @CC_LG
                              3011XC2
                              218,108 @*)
                           end
                           return items := `1;
                        end;

                        query GetOverdueBorrowers( out borrowers : Borrower [0:n]) =
                           begin
                              opt_set `2 : Borrower;
                              (* @CC_LG 6010(96,56)(97,29) @*)
                              => `2 : with (    -borrows->
                                             &  overdue ( 100 ) );
                              (* @CC_LG
                              3011XC2
                              218,108 @*)
                           end
                           return borrowers := `2;
                        end;

                        query GetBorrowerForItem( item : Item ; out borrower : Borrower) =
                           begin
                              obl_node `1 = item;
                              (* @CC_LG 6007(96,56)(97,29) @*)
                              obl_node `2 : Borrower;
                              (* @CC_LG 6007(96,154)(97,29) @*)
                              `2 -> `1 : borrows;
                              (* @CC_LG 3004C @*)
                           end
                           return borrower := `2;
                        end;

                        query GetItemsForBorrower( borrower : Borrower ; out items : Item [0:n]) =
                           begin
                              obl_node `1 = borrower;
                              (* @CC_LG 6007(96,154)(97,29) @*)
                              opt_set `2 : Item;
                              (* @CC_LG 6010(96,56)(97,29) @*)
                              `1 -> `2 : borrows;
                              (* @CC_LG 3004C @*)
                           end
                           return items := `2;
                        end;

                        query ReliableBorrower( borrower : Borrower ; library : LIBRARY) =
                           begin
                              obl_node `1 = borrower;
                              (* @CC_LG 6007(96,154)(97,29) @*)
                              not_node `2 : Item;
                              (* @CC_LG 6013(96,56)(97,29) @*)
                              `1 => `2 : hasItem ( library );
                              (* @CC_LG 3008C @*)
                              => `2 : overdue ( 100 );
                              (* @CC_LG
                              3011XC2
                              218,108 @*)
                           end
                        end;

                  end;

                  section LIB_LENDING_OPERATIONS

                     declares

                        transformation LendOutItem ( borrower : Borrower ; item : Item) =

                           begin
                              obl_node `1 = borrower;
                              (* @CC_LG 6007(360,154)(97,29) @*)
                              obl_node `2 = item;
                              (* @CC_LG 6007(216,56)(97,29) @*)
                              obl_node `3 : LendingLibrary;
                              (* @CC_LG 6007(216,154)(97,29) @*)
                              `3 -> `2 : owns;
                              (* @CC_LG 3004C @*)
                              not_node `4 : Borrower;
                              (* @CC_LG 6013(72,154)(97,29) @*)
                              `4 -> `2 : borrows;
                              (* @CC_LG 3004C @*)
                              `1 +> `2 : borrows;
                              (* @CC_LG 3010C @*)
                           end
                           ::=
                           begin
                              obl_node 1' = `1;
                              (* @CC_LG 6007(96,154)(97,29) @*)
                              obl_node 2' = `2;
                              (* @CC_LG 6007(168,56)(97,29) @*)
                              obl_node 3' = `3;
                              (* @CC_LG 6007(240,154)(97,29) @*)
                              3' -> 2' : owns;
                              (* @CC_LG 3004C @*)
                              1' -> 2' : borrows;
                              (* @CC_LG 3004C @*)
                           end
                           condition card ( `1.hasItem ( `3 ) ) <= `3.bookLimit;
                           transfer 2'.loanDate := CurrentDate;
                        end;

                        transformation LendOutItems ( borrower : Borrower ; item : Item [1:n]) =

                           begin
                              obl_node `1 = borrower;
                              (* @CC_LG 6007(96,154)(97,29) @*)
                              obl_set `2 = item;
                              (* @CC_LG 6011(168,56)(97,29) @*)
                              obl_node `3 : LendingLibrary;
                              (* @CC_LG 6007(240,154)(97,29) @*)
                              `3 -> `2 : owns;
                              (* @CC_LG 3004C @*)
                              `1 +> `2 : borrows;
                              (* @CC_LG 3010C @*)
                              => `2 : not with <-borrows-;
                              (* @CC_LG
                              3011XC2
                              290,108 @*)
                           end
                           ::=
                           begin
                              obl_node 1' = `1;
                              (* @CC_LG 6007(96,154)(97,29) @*)
                              obl_set 2' = `2;
                              (* @CC_LG 6011(168,56)(97,29) @*)
                              obl_node 3' = `3;
                              (* @CC_LG 6007(240,154)(97,29) @*)
                              3' -> 2' : owns;
                              (* @CC_LG 3004C @*)
                              1' -> 2' : borrows;
                              (* @CC_LG 3004C @*)
                           end
                           condition card ( `1.hasItem ( `3 ) ) <= `3.bookLimit;
                           transfer 2'.loanDate := CurrentDate;
                        end;

                        transformation ReturnItems ( borrower : Borrower ; items : Item [1:n]) =

                           begin
                              obl_node `1 = borrower;
                              (* @CC_LG 6007(240,154)(97,29) @*)
                              obl_set `2 = items;
                              (* @CC_LG 6011(168,56)(97,29) @*)
                              obl_node `3 : LendingLibrary;
                              (* @CC_LG 6007(96,154)(97,29) @*)
                              `3 -> `2 : owns;
                              (* @CC_LG 3004C @*)
                              `1 -> `2 : borrows;
                              (* @CC_LG 3004C @*)
                           end
                           ::=
                           begin
                              obl_node 1' = `1;
                              (* @CC_LG 6007(240,56)(97,29) @*)
                              obl_set 2' = `2;
                              (* @CC_LG 6011(96,56)(97,29) @*)
                              obl_node 3' = `3;
                              (* @CC_LG 6007(96,154)(97,29) @*)
                              3' -> 2' : owns;
                              (* @CC_LG 3004C @*)
                           end
                           condition card ( `1.hasItem ( `3 ) ) <= `3.bookLimit;
                           transfer 2'.loanDate := nil;
                        end;

                  end;

            end;

            section LIB_USER_QUERIES_AND_RESERVATIONS

               declares

                  section LIB_USER_QUERIES

                     declares

                        query FindItemsToBorrow( titleSubstring : string ; out item : Item [1:n]) =
                           begin
                              obl_set `1 : Item;
                              (* @CC_LG 6011(96,158)(97,29) @*)
                              obl_node `2 : TITLE;
                              (* @CC_LG 6007(96,56)(97,29) @*)
                              `1 -> `2 : has;
                              (* @CC_LG 3004C @*)
                              => `2 : valid hasSubstring ( self.title, titleSubstring );
                              (* @CC_LG
                              3011XC2
                              218,108 @*)
                              => `1 : valid ((self.<-owns- : LendingLibrary).bookLimit > 0);
                              (* @CC_LG
                              3011XC2
                              218,210 @*)
                           end
                           return item := `1;
                        end;

                        query FindOneItemToLookup( titleSubstring : string ; out items : Item) [0:n] =
                           begin
                              obl_node `1 : ReferenceLibrary;
                              (* @CC_LG 6007(96,252)(97,29) @*)
                              obl_node `2 : Item;
                              (* @CC_LG 6007(96,154)(97,29) @*)
                              obl_node `3 : TITLE;
                              (* @CC_LG 6007(96,56)(97,29) @*)
                              `1 -> `2 : owns;
                              (* @CC_LG 3004C @*)
                              `2 -> `3 : has;
                              (* @CC_LG 3004C @*)
                              => `3 : valid hasSubstring ( self.title, titleSubstring );
                              (* @CC_LG
                              3011XC2
                              218,108 @*)
                           end
                           return items := `2;
                        end;

                        function hasSubstring : ( s : string ; match : string) -> boolean =
                           exist i := elem ( 1 .. (length ( s ) - length ( match ) + 1) ) ::
                              use subStr := substr ( s, i, i + length ( match ) - 1 ) ::
                                 match = subStr
                              end
                           end
                        end;

                        query FindAllItems( titleSubstring : string ; out items : Item [0:n]) =
                           use t : TITLE [1:n];
                               i : Item [1:n]
                           do
                              FindTitle ( titleSubstring, out t )
                              & FindItemsForTitle ( t, out i )
                              & items := i
                           end
                        end;

                        query FindAllAvailableItems( titleSubstring : string ; out items : Item [0:n])
                        =
                           use t : TITLE [1:n];
                               i : Item [1:n]
                           do
                              FindTitle ( titleSubstring, out t )
                              & FindItemsForAllTitles ( t, out i )
                              & CheckAvailability ( i, out i )
                              & items := i
                           end
                        end;

                        query FindTitle( titleSubstring : string ; out titles : TITLE [1:n]) =
                           begin
                              obl_set `1 : TITLE;
                              (* @CC_LG 6011(96,56)(97,29) @*)
                              => `1 : valid hasSubstring ( self.title, titleSubstring );
                              (* @CC_LG
                              3011XC2
                              218,108 @*)
                           end
                           return titles := `1;
                        end;

                        query FindItemsForAllTitles( titles : TITLE [1:n] ; out items : Item [1:n]) =
                           use i, all_i : Item [0:n]
                           do
                              all_i := nil
                              & for_all t := elem ( titles )
                                do
                                   FindItemsForTitle ( t, out i )
                                   & all_i := (all_i or i)
                                end
                              & items := all_i
                           end
                        end;

                        query FindItemsForTitle( titles : TITLE ; out items : Item [1:n]) =
                           begin
                              obl_node `1 = titles;
                              (* @CC_LG 6007(96,56)(97,29) @*)
                              obl_set `2 : Item;
                              (* @CC_LG 6011(96,158)(97,29) @*)
                              `2 -> `1 : has;
                              (* @CC_LG 3004C @*)
                           end
                           return items := `2;
                        end;

                        query CheckAvailability( inItems : Item [1:n] ; out outItems : Item [1:n]) =
                           begin
                              obl_set `1 = inItems;
                              (* @CC_LG 6011(96,56)(97,29) @*)
                              => `1 : not with <-borrows-;
                              (* @CC_LG
                              3011XC2
                              218,108 @*)
                              => `1 : valid (self.<-owns-.bookLimit > 0);
                              (* @CC_LG
                              3011XC2
                              218,108 @*)
                           end
                           return outItems := `1;
                        end;

                  end;

            end;

            section INIT_LIBRARY

               declares

                  transformation CreateLIS =

                     ::=
                     begin
                        1' : LIS;
                        (* @CC_LG 6012(96,56)(97,29) @*)
                     end
                  end;

                  transformation AddLibrary ( libType : type_in LIBRARY ; out library : libType) [1:1]
                     =

                     begin
                        obl_node `1 : LIS;
                        (* @CC_LG 6007(96,56)(97,29) @*)
                     end
                     ::=
                     begin
                        2' : libType;
                        (* @CC_LG 6012(96,154)(97,29) @*)
                        obl_node 1' = `1;
                        (* @CC_LG 6007(96,56)(97,29) @*)
                        2' -> 1' : partOf;
                        (* @CC_LG 3004C @*)
                     end
                     return library := 2';
                  end;

                  transformation InitLIS =

                     use t1, t2, t3 : TITLE;
                         l, r : LIBRARY;
                         i1, i2, i3 : Item;
                         b1, b2, b3 : Borrower
                     do
                        CreateLIS
                        & AddLibrary ( LendingLibrary, out l )
                        & AddLibrary ( ReferenceLibrary, out r )
                        & skip
                        & AddTitle ( Book, "Vol.1: Foundations", "G. Rozenberg", out t1 )
                        & AddTitle ( Book, "Vol.2: Specifications ...", "G. Rozenberg", out t2 )
                        & AddTitle ( Series, "Handbook of Graph Grammars", "G. Rozenberg", out t3 )
                        & MakeSubtitle ( t3, t2 )
                        & MakeSubtitle ( t3, t1 )
                        & AddItem ( t1, l, out i1 )
                        & AddItem ( t2, l, out i2 )
                        & AddItem ( t1, r, out i1 )
                        & AddItem ( t2, r, out i2 )
                        & AddTitle ( Book, "LaTeX User's Guide", "L. Lamport", out t1 )
                        & AddItem ( t1, l, out i1 )
                        & AddItem ( t1, l, out i2 )
                        & AddItem ( t1, r, out i3 )
                        & AddTitle ( Book, "Java in a Nutshell", "D. Flanagan", out t1 )
                        & AddItem ( t1, l, out i1 )
                        & AddTitle ( Book, "UML Distilled", "M. Fowler", out t1 )
                        & AddItem ( t1, l, out i1 )
                        & AddTitle ( Book, "The IPSEN Approach", "M. Nagl", out t1 )
                        & AddItem ( t1, l, out i1 )
                        & AddTitle ( Book, "Object-Oriented Software Construction", "B. Meyer", out t1 )
                        & AddItem ( t1, l, out i1 )
                        & AddTitle ( Book, "Sommertime in the Big Woods", "L. Ingalls Wilder", out t1 )
                        & AddItem ( t1, l, out i1 )
                        & AddItem ( t1, r, out i2 )
                        & AddTitle ( Book, "Alice in Wonderland", "L. Carroll", out t1 )
                        & AddItem ( t1, l, out i1 )
                        & AddItem ( t1, l, out i2 )
                        & AddItem ( t1, l, out i3 )
                        & AddTitle ( Book, "Frankenstein", "M.Shelley", out t1 )
                        & AddItem ( t1, l, out i1 )
                        & AddItem ( t1, l, out i2 )
                        & skip
                        & AddLibUser ( l, "Hannah", out b1 )
                        & AddLibUser ( l, "Timo", out b2 )
                        & AddLibUser ( l, "Andy", out b3 )
                     end
                  end;

                  transformation Example =

                     use t1, t2, t3 : TITLE;
                         t : TITLE [0:n];
                         l, r : LIBRARY;
                         i1, i2, i3 : Item;
                         b1, b2, b3 : Borrower
                     do
                        CreateLIS
                        & AddLibrary ( LendingLibrary, out l )
                        & AddLibrary ( ReferenceLibrary, out r )
                        & AddLibUser ( l, "Hannah", out b1 )
                        & AddLibUser ( l, "Timo", out b2 )
                        & AddLibUser ( l, "Andy", out b3 )
                        & AddTitle ( Book, "Vol.1: Foundations", "G. Rozenberg", out t1 )
                        & AddTitle ( Book, "Vol.2: Specifications ...", "G. Rozenberg", out t2 )
                        & AddTitle ( Series, "Handbook of Graph Grammars", "G. Rozenberg", out t3 )
                        & MakeSubtitle ( t3, t2 )
                        & MakeSubtitle ( t3, t1 )
                        & AddItem ( t1, l, out i1 )
                        & LendOutItem ( b3, i1 )
                        & AddItem ( t2, l, out i2 )
                        & LendOutItem ( b1, i2 )
                        & AddItem ( t1, r, out i1 )
                        & AddItem ( t2, r, out i2 )
                        & AddTitle ( Book, "LaTeX User's Guide", "L. Lamport", out t1 )
                        & AddItem ( t1, l, out i1 )
                        & LendOutItem ( b1, i1 )
                        & AddItem ( t1, l, out i2 )
                        & AddItem ( t1, r, out i3 )
                        & AddTitle ( Book, "Java in a Nutshell", "D. Flanagan", out t1 )
                        & AddItem ( t1, l, out i1 )
                        & AddTitle ( Book, "UML Distilled", "M. Fowler", out t1 )
                        & AddItem ( t1, l, out i1 )
                        & LendOutItem ( b1, i1 )
                        & AddTitle ( Book, "The IPSEN Approach", "M. Nagl", out t1 )
                        & AddItem ( t1, l, out i1 )
                        & AddTitle ( Book, "Object-Oriented Software Construction", "B. Meyer", out t1 )
                        & AddItem ( t1, l, out i1 )
                        & AddTitle ( Book, "Sommertime in the Big Woods", "L. Ingalls Wilder", out t1 )
                        & AddItem ( t1, l, out i1 )
                        & AddItem ( t1, r, out i2 )
                        & AddTitle ( Book, "Alice in Wonderland", "L. Carroll", out t1 )
                        & AddItem ( t1, l, out i1 )
                        & AddItem ( t1, l, out i2 )
                        & AddItem ( t1, l, out i3 )
                        & AddTitle ( Book, "Frankenstein", "M.Shelley", out t1 )
                        & AddItem ( t1, l, out i1 )
                        & AddItem ( t1, l, out i2 )
                        & LendOutItem ( b2, i1 )
                        & FindTitle ( "Software", out t )
                        & FindTitle ( "Construction", out t )
                        & skip
                     end
                  end;

            end;

            section LIB_MAIN

               declares

                  transformation Main =

                     Example
                  end;

            end;

      end;

end.
(* @CC_LG 7012(12,8)(1029,515) @*)
