/**************/
/* DateTime.c */
/**************/

/**************************************************/
/* implementation of imported types and functions */
/**************************************************/

/***************************************************************************/
/*                                                                         */
/*  ATTENTION: please SET environment Variable "IMPORTLIB" according       */
/*             to the library containing the object code of the            */
/*             imported types and functions.                               */
/*                                                                         */
/***************************************************************************/

/* Note further that an imported type needs to be provided with at 
 * least three pre-defined functions that are needed for PROGRES built-in 
 * functionality.
 * (let USERTYPE be the type to be imported)
 *  - USERTYPE_Size returns the size of the type in the parameter
 *  - USERTYPE_ValueToString writes the contents of the imported type 
 *    to a string, that will be returned in the first parameter
 *  - USERTYPE_StringToValue converts a string to USERTYPE and returns it by
 *    parameter
 * Remark: The formerly needed procedures WriteConst and ReadConst 
 * are obsolete!

 * Further user-defined functions for an imported type have the following 
 * rules: 
 * Assume that you want to import a function in the specification like
 *   MyImportedFunction : (T1, ..., Tn) -> Tr
 * where types T1, ..., Tn, Tr are imported or built-in types.
 * The procedure implementing the function above must then look like:
 *   PROCEDURE MyImportedFunction (Tr *pr, T1 *p1, ..., Tn *pn);
 * p1, ..., pn are the in-paramters and pr is the result of the function.
 *
 * Note: All parameters must be passed by address (for technical reasons)
 */


#include <stdio.h>
#include <stdlib.h>

/* use type definition of text in the header file */
#include "DateTime.h"


/***********************************************************/
/* implementation of standard functions for imported types */
/***********************************************************/

/* The following three procedures are obligatory in connection 
 * with an imported type. They must not be used in the specification
 * explicitely, but they are needed for internal use (e.g. conversion
 * from text to string and vice versa. 
 */
/* Note: The procedure names and the parameters (type and order) 
 * are important. The convention is, that in procedures results are 
 * returned as the first parameter. 
 */

void Date_Size (int *size)
/* tell only the size of the type */
{
  *size = sizeof (Date);
}


void Date_ValueToString (PGCMString *outString, Date *value)
/* convert Date to string */
{
  sprintf((*outString).s, "%d.%d.%d", value->day, value->month, value->year);
}


void Date_StringToValue (Date *value, PGCMString *inString)
/* convert string to Date */
{
  sscanf((*inString).s, "%d.%d.%d", &(value->day), &(value->month), &(value->year));
}

void Time_Size( int* size )
{
  *size = sizeof (Time);
}

void Time_ValueToString (PGCMString *outString, Time *value)
/* Time always looks like "xx:yy:zz" */
{
  sprintf((*outString).s, "%2d:%2d:%2d", value->hour, value->min, value->sec);
  /* add some leading zeros - yes, it's a little bit hacked... */
  if (value->hour < 10) {
    (*outString).s[0] = '0';
  }
  if (value->min < 10) {
    (*outString).s[3] = '0';
  }
  if (value->sec < 10) {
    (*outString).s[6] = '0';
  }
}

void Time_StringToValue (Time *value, PGCMString *inString)
/* the input string must be conform to the "xx:yy:zz" format! */
{
  sscanf((*inString).s, "%2d:%2d:%2d", &(value->hour),
	 &(value->min), &(value->sec));
}

/*****************************************************/
/* implementation of user-defined imported functions */
/*****************************************************/

void CurrentDate (Date *out)
{
  time_t curr;
  struct tm *ptime;

  curr = time(&curr);
  ptime = localtime(&curr);
  sprintf(out->dateString, "%s", asctime(ptime));
  out->day = ptime->tm_mday;
  out->month = ptime->tm_mon + 1;
  out->year = ptime->tm_year + 1900;
  out->inttime = curr;
}

void AddDays (Date *out, Date *in1, int *days )
{
  int limit[12] = {31,28,31,30,31,30,31,31,30,31,30,31};
    
  out->day = in1->day + (*days);
  out->month = in1->month;
  out->year = in1->year;
  while (out->day > limit[out->month - 1]) {
    if (((out->year % 4) == 0) && ((out->year % 400) != 0)) {
      limit[1] = 29;
    } else {
      limit[1] = 28;
    }
    out->day -= limit[out->month - 1];
    out->month += 1;
    if (out->month > 12) {
      out->month -= 12;
      out->year += 1;
    }
  }
}

void Days ( int *outVal, Date *date2, Date *date1 )
{
  *outVal = (int) (difftime(date2->inttime, date1->inttime)/(60*60*24));
}
 
void IsToday ( BOOLEAN *s, Date *date )
{
  Date *tmp;
  
  CurrentDate(tmp);
  if ((date->day == tmp->day) && (date->month == tmp->month) &&
      (date->year == tmp->year)) {
    *s = TRUE;
  } else {
    *s = FALSE;
  }
}

void CurrentTime (Time *out)
{
  time_t curr;
  struct tm *ptime;

  curr = time(&curr);
  ptime = localtime(&curr);
  sprintf(out->dateString, "%s", asctime(ptime));
  out->hour = ptime->tm_hour;
  out->min = ptime->tm_min;
  out->sec = ptime->tm_sec;
  out->inttime = curr;
}

void DiffTime (Time *out, Time *later, Time *earlier)
/* computes the difference between the later time and the
   earlier time */
{
  out->hour = (later->hour - earlier->hour);
  out->min = (later->min - earlier->min);
  out->sec = (later->sec - earlier->sec);
  if (out->sec<0) {
    out->sec += 60;
    out->min -= 1;
  }
  if (out->min<0) {
    out->min += 60;
    out->hour -= 1;
  }
}

void AddTime (Time *out, Time *in1, Time *in2)
{
  out->hour = in1->hour + in2->hour;
  out->min = in1->min + in2->min;
  out->sec = in1->sec + in2->sec;
  if (out->sec > 59){
    out->sec -= 60;
    out->min += 1;
  }
  if (out->min > 59){
    out->min -= 60;
    out->hour += 1;
  }
}

  
